/* -*- Mode: C; c-basic-offset:4 ; -*- */
/*
 *  (C) 2001 by Argonne National Laboratory.
 *      See COPYRIGHT in top-level directory.
 */

#include "mpidimpl.h"

/* FIXME: Explain this function */

/* FIXME: should there be a simpler version of this for short messages, 
   particularly contiguous ones?  See also the FIXME about buffering
   short messages */

#undef FUNCNAME
#define FUNCNAME MPIDI_Isend_self
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
#if defined(FINEGRAIN_MPI)
int MPIDI_Isend_self(const void ** buf_handle, int count, MPI_Datatype datatype, int rank, int tag, MPID_Comm * comm, int context_offset,
		     int type, MPID_Request ** request)
#else
int MPIDI_Isend_self(const void * buf, int count, MPI_Datatype datatype, int rank, int tag, MPID_Comm * comm, int context_offset,
		     int type, MPID_Request ** request)
#endif
{
    MPIDI_Message_match match;
    MPID_Request * sreq;
    MPID_Request * rreq;
    MPIDI_VC_t * vc;
#if defined(MPID_USE_SEQUENCE_NUMBERS)
    MPID_Seqnum_t seqnum;
#endif    
    int found;
    int mpi_errno = MPI_SUCCESS;

    FGP_DBG_ENTER(MPIDI_Isend_self);
    MPIU_DBG_MSG(CH3_OTHER,VERBOSE,"sending message to self");

#if defined(FINEGRAIN_MPI)
    void * buf = (void *) (*buf_handle);
    int destpid=-1, destworldrank=-1;
    MPIDI_Comm_get_pid_worldrank(comm, rank, &destpid, &destworldrank);
#endif
    
    /* HK:
       MPIDI_Request_create_sreq() macro uses MPID_Request_mem is global and shared by FGPs.
       HK: TODO Double-check **rank in request macro** 
       Note: In MPIDI_Request_create_sreq() "sreq" maintains information about who to send to
       in (sreq_)->dev.match.dest_rank = rank; where rank was passed as parameter to MPID_Send or MPID_Isend.
       In "sreq", it has not replaced that with its own(i.e. source's rank). That replacement is done when
       filling in the match struct.
    */    
    MPIDI_Request_create_sreq(sreq, mpi_errno, goto fn_exit); 
    MPIDI_Request_set_type(sreq, type);
    MPIDI_Request_set_msg_type(sreq, MPIDI_REQUEST_SELF_MSG);
#if defined(FINEGRAIN_MPI)
    if ( MPIDI_REQUEST_TYPE_ZSEND == type ){
        MPIDI_Request_set_self_zerocopy_flag(sreq, TRUE);
        sreq->dev.user_buf_handle  = (void **)buf_handle;
        sreq->dev.user_buf = NULL;

    }
#endif
    /* HK:
       Extending send_self so that it would also handle communication between FGPs within same HWP.
       Previously, the following code was used (i.e. match.rank = rank), because in case of send_self,
       the source and destination is same. and the rank passed to MPI_Send() was fine. 
       Previous code:-
       match.rank = rank;
       #if defined(FINEGRAIN_MPI)
       match.dest_rank = rank; FGP MATCH This is the rank passed to MPI_Send() and is fine.
       #endif
       
       However, since we are extending it to handle communication between FGPs within same HWP, we need
       to have the sender's fgrank in the match struct. This is now done using ASSIGN_MATCH_RANK below.
    */
    ASSIGN_MATCH_RANK(match, comm, destworldrank);
    match.tag = tag;
    match.context_id = comm->context_id + context_offset;
    rreq = MPIDI_CH3U_Recvq_FDP_or_AEU(&match, &found); /* HK: matching now incorporates the destination fgrank. */
    FGP_DBG_PRINTF("After MPIDI_CH3U_Recvq_FDP_or_AEU found=%d\n");
    /* --BEGIN ERROR HANDLING-- */
    if (rreq == NULL)
    {
	MPIU_Object_set_ref(sreq, 0);
        if ( MPIDI_Request_get_self_zerocopy_flag(sreq) ){
            /* Free the MPIX_Zsend/Izsend sender's buffer */
            MPIU_Free(buf);
        }
	MPIDI_CH3_Request_destroy(sreq); 
	sreq = NULL;
        MPIU_ERR_SET1(mpi_errno, MPI_ERR_NO_MEM, "**nomem", "**nomemuereq %d", MPIDI_CH3U_Recvq_count_unexp());
	goto fn_exit;
    }
    /* --END ERROR HANDLING-- */

    /* HK: MPIDI_Comm_get_vc defined for FINEGRAIN_MPI
       HK: TODO IMPORTANT: Double-check that where ever MPIDI_Comm_get_vc
       is called in the code, it is passed the fgrank from MPI API routines
       that use fgrank.
    */
#if defined(FINEGRAIN_MPI)
    MPIDI_Comm_get_vc_direct(comm, destpid, &vc);
#else
    MPIDI_Comm_get_vc(comm, rank, &vc);
#endif
    MPIDI_VC_FAI_send_seqnum(vc, seqnum);
    MPIDI_Request_set_seqnum(sreq, seqnum);
    MPIDI_Request_set_seqnum(rreq, seqnum);

#if defined(FINEGRAIN_MPI)
    rreq->status.MPI_SOURCE = match.rank; /* HK: See comment above ASSIGN_MATCH_RANK. */
#else
    rreq->status.MPI_SOURCE = rank; 
#endif    
    rreq->status.MPI_TAG = tag;
    
    if (found)
    {
	MPIDI_msg_sz_t data_sz;

#if defined(FINEGRAIN_MPI)
        /*
          Pre-posted receive request found. Checking for different sender-receiver
          match pairing scenarios.
        */
        if ( MPIDI_Request_get_self_zerocopy_flag(sreq) && MPIDI_Request_get_self_zerocopy_flag(rreq) )
        {
            int sdt_contig;
            MPI_Aint sdt_true_lb;
            MPID_Datatype * sdt_ptr;
            
            FGP_DBG_PRINTF("%d: Preposted Recv-Collocated MPIX_Zsend/Izsend - MPIX_Zrecv/Izrecv pairing\n", my_fgrank); /* HK: FG_ZEROCOPY matched pair */
            assert(NULL == rreq->dev.user_buf);
            *(rreq->dev.user_buf_handle) = (void*) (*buf_handle);

            MPIDI_Datatype_get_info(count, datatype, sdt_contig, data_sz, sdt_ptr, sdt_true_lb);
            
            /* MPIX_Zsend buf_handle can't be set to NULL as we don't have
               a ptr to void**. */
        }
        else if( MPIDI_Request_get_self_zerocopy_flag(sreq) && !MPIDI_Request_get_self_zerocopy_flag(rreq) ){
            FGP_DBG_PRINTF("%d: Preposted Recv-Collocated MPIX_Zsend/Izsend - MPI_Recv/Irecv pairing. Freeing Sender's buffer\n", my_fgrank); /* HK: FG_ZEROCOPY freeing sender's buffer */
            MPIDI_CH3U_Buffer_copy(buf, count, datatype, &sreq->status.MPI_ERROR,
                                   rreq->dev.user_buf, rreq->dev.user_count, rreq->dev.datatype, &data_sz, &rreq->status.MPI_ERROR);
            /* Free the sender's buffer */
            MPIU_Free(buf);
        }
        else if( !MPIDI_Request_get_self_zerocopy_flag(sreq) && MPIDI_Request_get_self_zerocopy_flag(rreq) ){
            FGP_DBG_PRINTF("%d: Preposted Recv-Collocated MPI_Send/Isend - MPIX_Zrecv/Izrecv pairing. Allocating receiver's buffer.\n", my_fgrank); /* HK: FG_ZEROCOPY allocating receiver's buffer */
            assert(NULL == rreq->dev.user_buf);
            /* Added checks for buffer count size as is done in MPIDI_CH3U_Buffer_copy() */
            MPIDI_CH3U_Buffer_allocate(buf, count, datatype, &sreq->status.MPI_ERROR,
                                       rreq->dev.user_buf_handle, rreq->dev.user_count, rreq->dev.datatype, &data_sz, &rreq->status.MPI_ERROR);
            MPIDI_CH3U_Buffer_copy(buf, count, datatype, &sreq->status.MPI_ERROR,
                                   *(rreq->dev.user_buf_handle), rreq->dev.user_count, rreq->dev.datatype, &data_sz, &rreq->status.MPI_ERROR);            
        } else {
            FGP_DBG_PRINTF("%d: Preposted Recv-Collocated MPI_Send/Isend - MPI_Recv/Irecv pairing\n", my_fgrank);
#endif /* matches #if defined(FINEGRAIN_MPI) */
            MPIU_DBG_MSG(CH3_OTHER,VERBOSE,
		     "found posted receive request; copying data");
            MPIDI_CH3U_Buffer_copy(buf, count, datatype, &sreq->status.MPI_ERROR,
                                   rreq->dev.user_buf, rreq->dev.user_count, rreq->dev.datatype, &data_sz, &rreq->status.MPI_ERROR);
#if defined(FINEGRAIN_MPI)
        }
#endif

	rreq->status.count = (int)data_sz;
	MPID_REQUEST_SET_COMPLETED(rreq); /* HK: Increments MPIDI_CH3I_progress_completion_count, which is now per FGP */
	MPID_Request_release(rreq); /* HK: Calls MPIDI_CH3_Request_destroy(rreq) TODO IMPORTANT Double-check FGP (also marked as TODO above) */
	/* sreq has never been seen by the user or outside this thread, so it is safe to reset ref_count and cc */
	MPIU_Object_set_ref(sreq, 1);
	sreq->cc = 0;
    }
    else
    {
	if (type != MPIDI_REQUEST_TYPE_RSEND)
	{
	    int dt_sz;
	
	    /* FIXME: Insert code here to buffer small sends in a temporary buffer? */

	    MPIU_DBG_MSG(CH3_OTHER,VERBOSE,
          "added receive request to unexpected queue; attaching send request");
	    if (HANDLE_GET_KIND(datatype) != HANDLE_KIND_BUILTIN)
	    {
		MPID_Datatype_get_ptr(datatype, sreq->dev.datatype_ptr);
		MPID_Datatype_add_ref(sreq->dev.datatype_ptr);
	    }
	    rreq->partner_request = sreq; /* HK: partner_request is set in MPIDI_Isend_self
                                             and used in MPIDI_CH3_RecvFromSelf */
	    rreq->dev.sender_req_id = sreq->handle;
	    MPID_Datatype_get_size_macro(datatype, dt_sz);
	    rreq->status.count = count * dt_sz;
	}
	else
	{
	    /* --BEGIN ERROR HANDLING-- */
	    MPIU_DBG_MSG(CH3_OTHER,TYPICAL,
			 "ready send unable to find matching recv req");
	    sreq->status.MPI_ERROR = MPIR_Err_create_code(MPI_SUCCESS, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER,
							  "**rsendnomatch", "**rsendnomatch %d %d", rank, tag);
	    rreq->status.MPI_ERROR = sreq->status.MPI_ERROR;
	    
	    rreq->partner_request = NULL;
	    rreq->dev.sender_req_id = MPI_REQUEST_NULL;
	    rreq->status.count = 0;
	    
	    /* sreq has never been seen by the user or outside this thread, so it is safe to reset ref_count and cc */
	    MPIU_Object_set_ref(sreq, 1);
	    sreq->cc = 0;
	    /* --END ERROR HANDLING-- */
	}
	    
	MPIDI_Request_set_msg_type(rreq, MPIDI_REQUEST_SELF_MSG);
	MPIDI_CH3_Progress_signal_completion(); /* HK: Increments MPIDI_CH3I_progress_completion_count */
    }

  fn_exit:
    
    FGP_DBG_EXIT(MPIDI_Isend_self);

    *request = sreq;
    return mpi_errno;
}
